from setuptools import setup

setup(
    name='key_store',
    version='0.2.0',
    description='Helper functions for getting objects easily from a list with multiple filter keys',
    url='https://github.com/rimaddo/key-store',
    maintainer='Rachel',
    maintainer_email='rachel.isabella.madden@gmail.com',
    license='MIT',
    packages=['key_store'],
    python_requires='>3.7.0',
)
